local BagPress = AceLibrary("AceAddon-2.0"):new("FuBarPlugin-2.0", "AceEvent-2.0", "AceComm-2.0", "AceDB-2.0", "AceConsole-2.0")

--------------------------
--   Library Settings   --
--------------------------

-- fubar
BagPress.cannotDetachTooltip = true
BagPress.defaultPosition = "RIGHT"
BagPress.hasIcon = true
BagPress.hasNoColor = true
BagPress.hideWithoutStandby = true
BagPress.independentProfile = true

-- acecomm
BagPress.OnCommReceive = { }
BagPress.commPrefix = "BagPress"


-------------------
--   Libraries   --
-------------------

local Tablet = AceLibrary("Tablet-2.0")
local DewDrop = AceLibrary("Dewdrop-2.0")


-------------------------
--   Local Functions   --
-------------------------

-- local Debug = function(...) BagPress:Print(format("Debug: %s", ...)) end


local function RegisterTradeEvents()
	BagPress:RegisterEvent("UI_ERROR_MESSAGE", "OnUIMessage")
	BagPress:RegisterEvent("UI_INFO_MESSAGE", "OnUIMessage")
	BagPress:RegisterEvent("TRADE_SHOW", "OnTradeShow")
end


local function UnregisterTradeEvents()
	BagPress:UnregisterEvent("UI_ERROR_MESSAGE")
	BagPress:UnregisterEvent("UI_INFO_MESSAGE")
	BagPress:UnregisterEvent("TRADE_SHOW")
end


local function CompressInit()
	if UnitIsPlayer("target") and not UnitIsUnit("target", "player") then
		if not CheckInteractDistance("target", 2) then
			BagPress:Print("Target is out of range")
			return
		end

		local n = UnitName("target")
		BagPress:Print("Initiating compression with "..n)
		local x = BagPress:SendCommMessage("WHISPER", n, "Compress_Init", nil)
		--BagPress:Print(tostring(x))
	else
		BagPress:Print("You must have a player other than yourself targeted to do that")
	end
end


local function GenerateItemTable(item_table, trade)
	if trade then
		item_table.empty = 0
		item_table.items = { }
		item_table.force = { }
	end

	for b = 0, 4 do -- bag number

		for s = 1, GetContainerNumSlots(b) do -- slot number
			local item_link = GetContainerItemLink(b, s)

			if not item_link then
				if trade then
					item_table.empty = item_table.empty + 1
				end

			else
				local tradable = true

				BagPressTooltip:SetOwner(UIParent, "ANCHOR_NONE")
				BagPressTooltip:SetBagItem(b, s)

				if BagPressTooltip:NumLines() > 1 and BagPressTooltipTextLeft2:GetText() then
					local t = BagPressTooltipTextLeft2:GetText()
					tradable = not (
						   t == ITEM_SOULBOUND
						or t == ITEM_BIND_ON_PICKUP
						or t == ITEM_BIND_QUEST
					)
				end

				BagPressTooltip:Hide()

				local item_id = select(2, strsplit(":", item_link))
				local max_stack = select(8, GetItemInfo(item_link))
				tradable = tradable and max_stack > 1

				if tradable then
					if trade then

						if BagPress.db.char.force.items[item_id] then
							tinsert(item_table.force, { b, s })

						else
							if not item_table.items[item_id] then
								item_table.items[item_id] = { }
								item_table.items[item_id].filter = BagPress.db.char.filter[item_id]
								item_table.items[item_id].max = max_stack
								item_table.items[item_id].count = max_stack
							end

							local curr_stack = select(2, GetContainerItemInfo(b, s))

							if curr_stack < item_table.items[item_id].count then
								item_table.items[item_id].count = curr_stack
								item_table.items[item_id].slot = { b, s }
							end

						end

					else
						item_table[item_id] = true

					end

				end
			end
		end
	end
	return item_table
end


local function SortByItemName(t, sorted, sortedNames)
	for key, val in pairs(t) do -- key = item id
		local iName = GetItemInfo(key)
		if not iName then
			iName = tostring(key)
		end
		sortedNames[key] = iName
		for i, val2 in ipairs(sorted) do
			if iName < sortedNames[val2] then
				tinsert(sorted, i, key)
				break
			end
			if i == #sorted then
				tinsert(sorted, key)
				break
			end
		end
		if #sorted == 0 then
			tinsert(sorted, key)
		end
	end

	return sorted
end

---------------------
--   Ace Methods   --
---------------------

function BagPress:OnInitialize()
	BagPress:RegisterDB("BagPressDB", "BagPressDBPC")

	BagPress:RegisterDefaults('char', {
		filter = { },
		force = {
			on = false,
			items = { },
		},
	} )

	BagPress.opts = {
		type = 'group',
		args = {
			compress = {
				type = 'execute',
				name = "Initiate inventory compression with target player",
				desc = "Initiate inventory compression with target player",
				func = function() CompressInit() end,
				guiHidden = true,
			},
			filter = {
				type = 'group',
				name = "Filter",
				desc = "Add or remove item filters",
				args = {
					list = {
						type = 'execute',
						name = "List",
						desc = "List current filters",
						func = function()
							local filter_keys = 0
							for key, val in pairs(self.db.char.filter) do
								local iName, _, iRarity = GetItemInfo(key)
								local reconstructed_link
								if iName then
									reconstructed_link = ITEM_QUALITY_COLORS[iRarity].hex.."|Hitem:"..key..":0:0:0:0:0:0:0|h["..iName.."]|h|r"
								else
									reconstructed_link = key
								end
								self:Print(reconstructed_link)
								filter_keys = filter_keys + 1
							end
							self:Print(filter_keys.." active filters found")
						end,
					},
					add = {
						type = 'text',
						validate = function(...)
							local item_id = select(2, strsplit(":", ...)) or ...
							return GetItemInfo(...) and not self.db.char.filter[item_id]
						end,
						error = "[%s] is not a valid item link or id.",
						get = false,
						set = function(...)
							local item_id = select(2, strsplit(":", ...)) or ...
							self.db.char.filter[item_id] = true
							self:Print(format("%s added to filter.", ...))
						end,
						usage = "<itemlink\124itemid>",
						message = "[%s] added to filtered items.",
						name = "Add Filter",
						desc = "Add an item to be filtered",
					},
					remove = {
						type = 'text',
						validate = function(...)
							local item_id = select(2, strsplit(":", ...)) or ...
							return item_id and self.db.char.filter[item_id]
						end,
						error = "[%s] is not a valid item link or id or is not being filtered.",
						get = false,
						set = function(...)
							local item_id = select(2, strsplit(":", ...)) or ...
							self.db.char.filter[item_id] = nil
							self:Print(format("%s removed from filter.", ...))
						end,
						usage = "<itemlink\124itemid>",
						message = "[%s] removed from filtered items.",
						name = "Remove Filter",
						desc = "Remove an item from item filter",
					},
				},
			},
			force = {
				type = 'group',
				name = "Force Trading",
				desc = "Options for items to force trade",
				args = {
					add = {
						type = 'text',
						validate = function(...)
							local item_id = select(2, strsplit(":", ...)) or ...
							return GetItemInfo(...) and not self.db.char.force.items[item_id]
						end,
						error = "[%s] is not a valid item link or id.",
						get = false,
						set = function(...)
							local item_id = select(2, strsplit(":", ...)) or ...
							self.db.char.force.items[item_id] = true
							self:Print(format("%s added to filter.", ...))
						end,
						usage = "<itemlink\124itemid>",
						message = "[%s] added to force-traded items.",
						name = "Add Forced Trade",
						desc = "Add an item to be force traded",
					},
					remove = {
						type = 'text',
						validate = function(...)
							local item_id = select(2, strsplit(":", ...)) or ...
							return item_id and self.db.char.force.items[item_id]
						end,
						error = "[%s] is not a valid item link or id or is not being filtered.",
						get = false,
						set = function(...)
							local item_id = select(2, strsplit(":", ...)) or ...
							self.db.char.force.items[item_id] = nil
							self:Print(format("%s removed from filter.", ...))
						end,
						usage = "<itemlink\124itemid>",
						message = "[%s] removed from force-traded items.",
						name = "Remove Forced Trade",
						desc = "Remove an item to be force traded",
					},
					list = {
						type = 'execute',
						name = "List",
						desc = "List current force-trade items",
						func = function()
							local filter_keys = 0
							for key, val in pairs(self.db.char.force.items) do
								local iName, _, iRarity = GetItemInfo(key)
								local reconstructed_link
								if iName then
									reconstructed_link = ITEM_QUALITY_COLORS[iRarity].hex.."|Hitem:"..key..":0:0:0:0:0:0:0|h["..iName.."]|h|r"
								else
									reconstructed_link = key
								end
								self:Print(reconstructed_link)
								filter_keys = filter_keys + 1
							end
							self:Print(filter_keys.." active filters found")
						end,
					},
				},
			},
		},
	}

	BagPress:RegisterChatCommand("/bagpress", "/bp", BagPress.opts)
	BagPress:SetCommPrefix(BagPress.commPrefix)

	-------------------------
	--   FuBar functions   --
	BagPress.OnTooltipUpdate = function()
		Tablet:SetHint("|cffeda55fClick|r to initiate inventory compression with target player.")
	end

	BagPress.OnClick = function()
		CompressInit()
	end

	---------------------
	--   Scan Tooltip   -
	CreateFrame("GameTooltip", "BagPressTooltip", UIParent, "GameTooltipTemplate")
	BagPressTooltip:SetWidth(1)
	BagPressTooltip:SetHeight(1)

end


function BagPress:OnEnable()
	BagPress:RegisterComm(BagPress.commPrefix, "WHISPER")
end


function BagPress:OnMenuRequest(level, value, inTooltip, value2)
	if level == 1 then
		DewDrop:AddLine(
			'text', "Filter",
			'hasArrow', true,
			'value', "filter",
			'tooltipTitle', "Filter",
			'tooltipText', "These items will NEVER be traded")
		DewDrop:AddLine(
			'text', "Forced-Trade Items",
			'hasArrow', true,
			'value', "force",
			'tooltipTitle', "Force-Trade Items",
			'tooltipText', "These items will ALWAYS be traded")
		DewDrop:AddSeparator()

	elseif level == 2 then
		if value == "filter" then
			local inventory = GenerateItemTable({ })
			local temp_menu = { }

			for key, val in pairs(inventory) do
				if self.db.char.filter[key] then
					temp_menu[key] = true
				else
					temp_menu[key] = false
				end
			end
			inventory = nil

			for key, val in pairs(self.db.char.filter) do
				if temp_menu[key] == nil then
					temp_menu[key] = true
				end
			end

			local menu_sorted = SortByItemName(temp_menu, { }, { })

			for i, item_id in ipairs(menu_sorted) do
				local iName, _, iRarity = GetItemInfo(item_id)
				local iText

				if not iName then
					iText = item_id
				else
					iText = ITEM_QUALITY_COLORS[iRarity].hex..iName
				end

				if not self.db.char.force.items[item_id] then
					DewDrop:AddLine(
					'text', iText,
					'checked', temp_menu[item_id],
					'func', function()
						if self.db.char.filter[item_id] then
							self.db.char.filter[item_id] = nil
						else
							self.db.char.filter[item_id] = true
						end
					end)
				end
			end

		-------------- force items
		elseif value == "force" then
			local inventory = GenerateItemTable({ })
			local temp_menu = { }

			for key, val in pairs(inventory) do
				if self.db.char.force.items[key] then
					temp_menu[key] = true
				else
					temp_menu[key] = false
				end
			end
			inventory = nil

			for key, val in pairs(self.db.char.force.items) do
				if temp_menu[key] == nil then
					temp_menu[key] = true
				end
			end

			local menu_sorted = SortByItemName(temp_menu, { }, { })

			for i, item_id in ipairs(menu_sorted) do
				local iName, _, iRarity = GetItemInfo(item_id)
				local iText

				if not iName then
					iText = item_id
				else
					iText = ITEM_QUALITY_COLORS[iRarity].hex..iName
				end

				if not self.db.char.filter[item_id] then
					DewDrop:AddLine(
					'text', iText,
					'checked', temp_menu[item_id],
					'func', function()
						if self.db.char.force.items[item_id] then
							self.db.char.force.items[item_id] = nil
						else
							self.db.char.force.items[item_id] = true
						end
					end)
				end
			end
		end
	end
end



----------------------
--   Trade Events   --
----------------------

local TradeThread
local OwnTradeItems

function BagPress:OnTradeShow()
	if not self:IsEventRegistered("BAG_UPDATE") then
		self:RegisterEvent("BAG_UPDATE", function()
		if type(TradeThread) == "thread" and coroutine.status(TradeThread) == "suspended" then
			coroutine.resume(TradeThread)
		end end)
	end

	---------------
	-- Coroutine --
	TradeThread = coroutine.create(function()
		-- Debug("started")
		for i, bagslot in ipairs(OwnTradeItems) do
			local b, s = bagslot[1], bagslot[2]
			while true do
				local locked = select(3, GetContainerItemInfo(b, s))
				if locked then
					-- Debug("paused")
					coroutine.yield()
					-- Debug("resumed")
				else
					break
				end
			end
			ClearCursor()
			PickupContainerItem(b, s)
			ClickTradeButton(i)
			ClearCursor()
		end

		OwnTradeItems = { }
		self:UnregisterEvent("BAG_UPDATE")

	end)
	--/Coroutine --
	---------------
	coroutine.resume(TradeThread)

end


function BagPress:OnUIMessage(...)
	if string.find(..., ERR_TRADE_CANCELLED) or string.find(..., ERR_TRADE_TOO_FAR) or string.find(..., ERR_TRADE_COMPLETE) then
		if type(TradeThread) == "thread" and coroutine.status(TradeThread) == "suspended" then
			self:UnregisterEvent("BAG_UPDATE")
			TradeThread = nil
		end
		OwnTradeItems = { }
		UnregisterTradeEvents()
		-- Debug("success")
	end

end


---------------------
--   Comm Events   --
---------------------

-- Recieved by Player 2
function BagPress.OnCommReceive:Compress_Init(prefix, sender, dist, ...)
	local OwnItems = GenerateItemTable({ }, true)

	BagPress:SendCommMessage("WHISPER", sender, "Compress_ReceivedPlayer2", OwnItems)

end


-- Recieved by Player 1
function BagPress.OnCommReceive:Compress_ReceivedPlayer2(prefix, sender, dist, player2_items)

	local player2_trade_items = { }

	OwnTradeItems = { }

	local OwnItems = GenerateItemTable({ }, true)

	local empty_diff = OwnItems.empty - player2_items.empty

	-- +1 when i add, -1 when i don't
	-- <=0 for me
	for index, val in pairs(player2_items.force) do
		tinsert(player2_trade_items, val)
		if #player2_trade_items == 6 or #player2_trade_items == OwnItems.empty then
			break
		end
	end

	for index, val in pairs(OwnItems.force) do
		tinsert(OwnTradeItems, val)
		if #OwnTradeItems == 6 or #OwnTradeItems == player2_items.empty then
			break
		end
	end

	for item_id, val in pairs(OwnItems.items) do

		if #OwnTradeItems == player2_items.empty or
		   #OwnTradeItems == 6 or
		   #player2_trade_items == OwnItems.empty or
		   #player2_trade_items == 6 then break
		end

		local p2item = player2_items.items[item_id]

		if p2item and (p2item.count + val.count <= val.max) then

			if empty_diff > 0 and not p2item.filter then
				tinsert(player2_trade_items, p2item.slot)
				empty_diff = empty_diff - 1

			elseif not val.filter then
				tinsert(OwnTradeItems, val.slot)
				empty_diff = empty_diff + 1

			end

		end
	end

	if #OwnTradeItems == 0 and #player2_trade_items == 0 then
		BagPress:Print("No tradable items")
		return

	else
		BagPress:SendCommMessage("WHISPER", sender, "Compress_TradeList", player2_trade_items)
	end
end


-- Recieved by Player 2
function BagPress.OnCommReceive:Compress_TradeList(prefix, sender, dist, trade_items)
	BagPress:Print("Compression request by "..sender)
	OwnTradeItems = trade_items
	RegisterTradeEvents()
	BagPress:SendCommMessage("WHISPER", sender, "Compress_StartTrade", nil)
end


-- Recieved by Player 1
function BagPress.OnCommReceive:Compress_StartTrade(prefix, sender, dist, message)
	if UnitName("target") == sender then
		RegisterTradeEvents()
		InitiateTrade("target")
	else
		BagPress:Print("Target lost, please try again")
	end
end
